#ifndef IncludedIntSet
#define IncludedIntSet

#define __USE_GNU
#include <string.h>

typedef
#ifdef WordSize64
#define LogWordSz 6
long
#else
#define LogWordSz 5
#endif
int word;

#define BitsInWord (1<<LogWordSz)

#define SetWordIndex(i) ((i)>>LogWordSz)
#define SetWord(s,i) (s)[SetWordIndex(i)]
#define SetMask(i) (1<<((i)&(BitsInWord-1)))
#define SetAdd(s,i) SetWord(s,i) |= SetMask(i)
#define SetDel(s,i) SetWord(s,i) &= ~SetMask(i)
#define Member(s,i) (SetWord(s,i) & SetMask(i))
#define SetVolume(n) (SetWordIndex((n)+BitsInWord-1))
#define SetExactCopy(dest,src,n) \
  (memcpy ((void *) (dest), (void *) (src), SetVolume ((n)) * sizeof (word)))
#define NewSet(n) ((word *) malloc (SetVolume((n)) * sizeof (word)))
#define NewEmptySet(n) ((word *) calloc (SetVolume((n)), sizeof (word)))
#define SetEmpty(s,n) (memset ((void *) s, 0, SetVolume((n)) * sizeof (word)))
#define SetFill(s,n) (memset ((void *) s, -1, SetVolume((n)) * sizeof (word)))

typedef word *IntSet;

inline static void SetUnion (IntSet dest, IntSet src1, IntSet src2, int n) {
  int i;
  for (i = 0; i < SetVolume (n); ++i) dest[i] = src1[i] | src2[i];
}

inline static
void SetIntersect (IntSet dest, IntSet src1, IntSet src2, int n) {
  int i;
  for (i = 0; i < SetVolume (n); ++i) dest[i] = src1[i] & src2[i];
}

inline static word SetMaskUpto (int i) {
  word m = SetMask (i-1);
  return m | (m-1);
}

inline static int Subset (IntSet s1, IntSet s2, int n) {
  int i;
  for (i = 0; i < SetWordIndex (n-1); ++i) if (s1[i] & ~s2[i]) return 0;
  return (s1[i] & ~s2[i] & SetMaskUpto (n)) == 0;
}

inline static
int SetIntersectEmpty (IntSet dest, IntSet src1, IntSet src2, int n) {
  int i;
  word w = 0;
  for (i = 0; i < SetWordIndex (n-1); ++i)
    w |= (dest[i] = src1[i] & src2[i]);
  w |= (dest[i] = src1[i] & src2[i]) & SetMaskUpto (n);
  return w == 0;
}

inline static int IsEmpty (IntSet s, int n) {
  int i;
  for (i = 0; i < SetWordIndex (n-1); ++i) if (s[i]) return 0;
  return (s[i] & SetMaskUpto (n)) == 0;
}

inline static int SetSizeBetween (IntSet s, int lo, int hi) {
  // unchecked indices
  int i, cnt = 0;
  word w;
  i = SetWordIndex (lo);
  w = s[i] & ~(SetMask (lo) -1);
  while (i < SetWordIndex (hi-1)) {
    while (w) ++cnt, w &= w-1;
    w = s[++i];
  }
  for (w &= SetMaskUpto (hi); w; ++cnt, w &= w-1);
  return cnt;
}

inline static int SetSize (IntSet s, int n) {
  return SetSizeBetween (s, 0, n);
}

inline static int SetFindNext (IntSet s, int n, int i) { // unchecked index
  int j = SetWordIndex (i);
  word w, m;
  i &= BitsInWord-1;
  m = ~((1<<i) -1);
  if (j < SetWordIndex (n)) {
    if ((w = m & s[j])) goto findBit;
    i = 0;
    while (++j < SetWordIndex (n))
      if ((w = s[j])) goto findBit;
    m = -1;
  }
  w = n & (BitsInWord-1);
  if (!(w = ((1<<w) -1) & m & s[j])) return n;
 findBit:
  while (!(w & (1<<i))) ++i;
  return (j<<LogWordSz) + i;
}

typedef IntSet BitMatrix;

#define BitMatrixVolume(m,n) (SetVolume ((n)) * (m))
#define NewEmptyBitMatrix(m,n) \
  (NewEmptySet ((((n)+BitsInWord-1)&~(BitsInWord-1)) * (m)))
#define BitMatrixRow(s,n,i) (&s[SetVolume((n)) * (i)])

#endif
