#ifndef IncludedObjList
#define IncludedObjList

#define New(n,t) ((t *) malloc ((n) * sizeof (t)))

typedef void *object;

typedef struct objList *ObjList;

struct objList {
  ObjList tail;
  object head;
};

inline static ObjList newListCell (object item) {
  ObjList p = New (1, struct objList);
  p->head = item;
  return p;
}

#define NewListCell(item) (newListCell ((object) (item)))

inline static ObjList cons (object item, ObjList t) {
  ObjList p = newListCell (item);
  p->tail = t;
  return p;
}

#define Cons(item,t) (cons ((object) (item), (t)))

inline static void consOn (ObjList *tp, object item) {
  *tp = cons (item, *tp);
}

#define ConsOn(t,item) (consOn (&(t), (object) (item)))

inline static ObjList *snoc (ObjList *tp, object item) {
  ObjList p = newListCell (item);
  *tp = p;
  return &p->tail;
}

#define Snoc(tp,item) (snoc ((tp), (object) (item)))

inline static int length (ObjList l) {
  int i;
  for (i = 0; l != NULL; ++i, l = l->tail);
  return i;
}

#define FreeObjList(l,freeHead) { \
  ObjList __p, __q; \
  for (__p = (l); __p; __p = __q) { \
    freeHead (__p->head); \
    __q = __p->tail; \
    free (__p); \
  } \
}

object *revlist2arrayN (ObjList l, int n) {
  object *a = New (n, object);
  while (n--) {
    ObjList t = l->tail;
    a[n] = l->head;
    free (l);
    l = t;
  }
  return a;
}

inline object *revlist2array (ObjList l) {
  return revlist2arrayN (l, length (l));
}

typedef struct DLobjList *DLObjList;

struct DLobjList {
  DLObjList tail;
  object head;
  DLObjList back;
};

inline static DLObjList newDLListCell (object item) {
  DLObjList p = New (1, struct DLobjList);
  p->head = item;
  return p;
}

inline static DLObjList startDLList (object item) {
  DLObjList p = newDLListCell (item);
  p->tail = p->back = NULL;
  return p;
}

#define StartDLList(item) (startDLList ((object) (item)))

inline static DLObjList appendDLListTail (DLObjList t, object item) {
  DLObjList p = newDLListCell (item);
  p->tail = NULL;
  p->back = t;
  t->tail = p;
  return p;
}

#define AppendDLListTail(t,item) (appendDLListTail ((t), (object) (item)))

inline static DLObjList prependDLList (DLObjList t, object item) {
  DLObjList p = newDLListCell (item);
  p->tail = t;
  p->back = NULL;
  t->back = p;
  return p;
}

#define PrependDLList(t,item) (prependDLList ((t), (object) (item)))

#endif
