#include <string.h>
#include "trie.h"

#define PrintableSetSize ('~'-' '+1)

inline static int allocTrie (ASCIITrie *root, char *s, int data) {
  ASCIITrie t = (ASCIITrie) malloc (sizeof (struct asciiTrie));
  t->prefix = s;
  t->branch = NULL;
  t->data = data;
  *root = t;
  return data;
}

inline static ASCIITrie *newTable () {
  return (ASCIITrie *) calloc (PrintableSetSize, sizeof (ASCIITrie));
}

int trieAdd (ASCIITrie *root, const char *s0, int newElement (const char *)) {
  ASCIITrie t = *root;
  const char *s = s0;
  if (t == NULL) return allocTrie (root, strdup (s), newElement (s0));
  for (;;) {
    char *p = t->prefix;
    while ((*p == *s) && (*p != '\0')) ++p, ++s;
    if (*p == '\0')
      if (*s == '\0') {
	int l = t->data;
	if (l == -1) t->data = l = newElement (s0);
	return l;
      }
      else {
	char c = *s++ - ' ';
	ASCIITrie *b = t->branch;
	if (b == NULL)
	  t->branch = b = newTable ();
	else if ((t = b[c]) != NULL) continue;
	return allocTrie (&b[c], strdup (s), newElement (s0));
      }
    else {
      ASCIITrie *b = newTable ();
      char c = *p - ' ';
      *p++ = '\0';
      allocTrie (&b[c], p, t->data);
      b[c]->branch = t->branch;
      t->branch = b;
      int l = newElement (s0);
      if (*s == '\0') t->data = l;
      else {
	c = *s++ - ' ';
	allocTrie (&b[c], strdup (s), l);
	t->data = -1;
      }
      return l;
    }
  }
}

int trieLookup (ASCIITrie t, const char *s) {
  int r;
  TrieLookup (t, *s, ++s, *s == '\0', r);
  return r;
}

  /*
int trieLookup (ASCIITrie t, const char *s) {
  ASCIITrie *b;
  for (;;) {
    if (t == NULL) return -1;
    char *p = t->prefix;
    while ((*p == *s) && (*p != '\0')) ++p, ++s;
    if (*s == '\0') return t->data;
    if (*p != '\0' || (b = t->branch) == NULL) return -1;
    t = b[*s++ - ' '];
  }
}
  */

static void traverseRec (ASCIITrie t, char *sbuf, int i,
			 void processElement (int, char *)) {
  int j;
  for (j = 0; sbuf[i] = t->prefix[j]; ++i, ++j);
  if (t->data != -1) {
    sbuf[i] = '\0';
    processElement (t->data, sbuf);
  }
  if (t->branch != NULL)
    for (j = 0; j < PrintableSetSize; ++j) {
      ASCIITrie t1 = t->branch[j];
      if (t1 != NULL) {
	sbuf[i] = j + ' ';
	traverseRec (t1, sbuf, i+1, processElement);
      }
    }
}

void traverseTrie (ASCIITrie t, char *sbuf, void processElement (int, char *)) {
  if (t != NULL) traverseRec (t, sbuf, 0, processElement);
}

void freeTrie (ASCIITrie t) {
  int j;
  if (t != NULL) {
    free (t->prefix);
    if (t->branch != NULL)
      for (j = 0; j < PrintableSetSize; ++j) freeTrie (t->branch[j]);
    free (t);
  }
}

static char **keys;
static int nIds;

void addKey (int id, char *key) {
  if (id < nIds) keys[id] = strdup (key);
}

char **flattenASCIITrie (ASCIITrie t, int n) {
  keys = (char **) calloc (n, sizeof (char *));
  if (keys != NULL) {
    nIds = n;
    traverseTrie (t, (char *) malloc (80), addKey); // temp -- FIX
  }
  return keys;
}
